@@ -48,8 +48,9 @@ namespace Banshee.Mirage
         private static string playlist_name = "Playlist Generator";
 
         public static List<DatabaseTrackInfo> seeds = new List<DatabaseTrackInfo>();
-        protected List<DatabaseTrackInfo> tracksOverride = new List<DatabaseTrackInfo>();
+        protected List<DatabaseTrackInfo> suggested = new List<DatabaseTrackInfo>();
         protected List<DatabaseTrackInfo> skipped = new List<DatabaseTrackInfo>();
+        protected List<DatabaseTrackInfo> played = new List<DatabaseTrackInfo>();
         protected DatabaseTrackInfo processed;
         protected Db db;
         InterfaceActionService action_service;
@@ -97,20 +98,33 @@ namespace Banshee.Mirage
                     AddinManager.CurrentLocalizer.GetString ("Remove all tracks from the play queue"),
                     OnClearPlaylist)
             );
+            /* TODO: Add Ban Button or "Exclude from Playlists"
+            action_service.GlobalActions.AddImportant (
+                new ActionEntry ("BanFromMiragePlaylistAction", Stock.No,
+                    AddinManager.CurrentLocalizer.GetString ("Ban Song"), null,
+                    AddinManager.CurrentLocalizer.GetString ("Permanently ban the selected song from generated playlists"),
+                    OnBanFromPlaylist)
+            );
+            */
+            action_service.GlobalActions.AddImportant (
+                new ActionEntry ("SaveMiragePlaylistAction", Stock.Add,
+                    AddinManager.CurrentLocalizer.GetString ("Save as Playlist"), null,
+                    AddinManager.CurrentLocalizer.GetString ("Save the Playlist to your Music Library"),
+                    OnSavePlaylist)
+            );
             
             action_service.GlobalActions.Add (new ToggleActionEntry [] {
                 new ToggleActionEntry ("ClearMiragePlaylistOnQuitAction", null, 
-                                       AddinManager.CurrentLocalizer.GetString ("Clear on Quit"), null, 
-                                       AddinManager.CurrentLocalizer.GetString ("Clear the play queue when quitting"), 
-                                       OnClearPlaylistOnQuit, 
-                                       MiragePlugin.ClearOnQuitSchema.Get ())
+                    AddinManager.CurrentLocalizer.GetString ("Clear on Quit"), null, 
+                    AddinManager.CurrentLocalizer.GetString ("Clear the play queue when quitting"), 
+                    OnClearPlaylistOnQuit, 
+                    MirageConfiguration.ClearOnQuitSchema.Get ())
             });
+
             
             ui_id = action_service.UIManager.AddUiFromResource ("GlobalUI.xml");
+            Properties.SetString ("ActiveSourceUIResource", "ActiveSourceUI.xml");
             Properties.SetString ("GtkActionPath", "/PlaylistContextMenu");
-                    
-            action_service.PlaybackActions["NextAction"].Activated += OnNextPrevAction;
-            action_service.PlaybackActions["PreviousAction"].Activated += OnNextPrevAction;
             
             ServiceManager.PlayerEngine.ConnectEvent (OnPlayerEvent, 
                                                       PlayerEvent.StartOfStream | 
@@ -129,7 +143,7 @@ namespace Banshee.Mirage
             action_service.GlobalActions.Remove ("ClearMiragePlaylistAction");
             action_service.GlobalActions.Remove ("ClearMiragePlaylistOnQuitAction");
             
-            if (MiragePlugin.ClearOnQuitSchema.Get ()) {
+            if (MirageConfiguration.ClearOnQuitSchema.Get ()) {
                 Clear ();
             }
         }
@@ -149,24 +163,10 @@ namespace Banshee.Mirage
                     VALUES (NULL, ?, -1, 0, 1)", playlist_name));
             }
         }
-        
-        public void FillPlaylist ()
-        {
-            // Add seed tracks 
-            lock (TrackModel) {
-                seeds.Clear();
-                seeds.AddRange (tracksOverride);
-                
-                tracksOverride.Clear();
-                
-                OnUpdated();
-            }
-            SimilarTracks (seeds, seeds);
-        }
-        
-        public delegate void UpdatePlaylistDelegate(int[] playlist, int length);
 
-        protected void UpdatePlaylist (int[] playlist, int length)
+        public delegate void UpdatePlaylistDelegate(int[] playlist, bool append);
+        
+        protected void UpdatePlaylist (int[] playlist, bool append)
         {
             Gtk.Application.Invoke(delegate {
                 if (playlist == null) {
@@ -174,12 +174,20 @@ namespace Banshee.Mirage
                     return;
                 }
                 
+                int length_wanted = MirageConfiguration.PlaylistLength.Get ();
                 lock (DatabaseTrackModel) {
-                    RemoveTrackRange (DatabaseTrackModel, new Hyena.Collections.RangeCollection.Range (current_track + 1, Count - 1));
+
+                    // check if we have to delete the suggested songs, or just
+                    // append
+                    if (!append) {
+                        RemoveTrackRange (DatabaseTrackModel, new Hyena.Collections.RangeCollection.Range (current_track+1, Count-1));
+                    }
+
                     int sameArtistCount = 0;
                     int i = 0;
                     int pi = 0;
-                    while ((i < Math.Min(playlist.Length, length)) && (pi < playlist.Length)) {
+                    suggested.Clear();
+                    while ((i < Math.Min(playlist.Length, length_wanted)) && (pi < playlist.Length)) {
                         DatabaseTrackInfo track = DatabaseTrackInfo.Provider.FetchSingle(playlist[pi]);
                         bool sameArtist = track.Artist.Equals(seeds[seeds.Count-1].Artist);
                         pi++;
@@ -187,16 +195,18 @@ namespace Banshee.Mirage
                         if (sameArtist)
                             sameArtistCount++;
                         
-                        if (sameArtist && (sameArtistCount > 0.5*length)) {
+                        // We don't want more than half of the tracks by the same artist
+                        if (sameArtist && (sameArtistCount > 0.5*length_wanted)) {
                             continue;
                         } else
                             i++;
-                        
+                       
+                        suggested.Add(track); 
                         AddTrack(track);
                     }
                     SetStatus (AddinManager.CurrentLocalizer.GetString ("Playlist ready."), false, false, null);
                 }
-                
+
                 OnUpdated();
                 HideStatus();
             });
@@ -207,14 +217,6 @@ namespace Banshee.Mirage
             // Reorder Tracks
         }*/
 
-        protected override void AddTrack(DatabaseTrackInfo track)
-        {
-            lock (DatabaseTrackModel) {
-                tracksOverride.Add(track);
-            }
-            base.AddTrack (track);
-        }
-        
         public override bool AcceptsInputFromSource (Source source)
         {
             return source == Parent || 
@@ -225,16 +227,51 @@ namespace Banshee.Mirage
             get { return SourceMergeType.ModelSelection; }
         }
         
-        public override void MergeSourceInput (Source source, SourceMergeType mergeType)
+        public override void MergeSourceInput (Source from, SourceMergeType mergeType)
         {
-            bool was_empty = (Count == 0);
-            base.MergeSourceInput (source, mergeType);
+            DatabaseSource source = from as DatabaseSource;
+            if (source == null || !(source.TrackModel is DatabaseTrackListModel)) {
+                return;
+            }
 
+            bool was_empty = (Count == 0);
+
+            // if playlist is empty start a new one
             if (was_empty) {
+                base.MergeSourceInput (source, mergeType);
+                seeds.Clear();
                 for (int i = 0; i < TrackModel.Count; i++) {
-                    tracksOverride.Add ((DatabaseTrackInfo)TrackModel[i]);
+                    seeds.Add ((DatabaseTrackInfo)TrackModel[i]);
                 }
-                FillPlaylist ();
+
+                played.Clear();
+                played.AddRange(seeds);
+
+                SimilarTracks (seeds, played, false);
+
+            // if playlist is not empty append the songs
+            } else {
+
+                seeds.Clear();
+                DatabaseTrackListModel tm = (DatabaseTrackListModel)(source.TrackModel);
+                CachedList<DatabaseTrackInfo> cached_list = CachedList<DatabaseTrackInfo>.CreateFromModelSelection (tm);
+                foreach (DatabaseTrackInfo ti in cached_list) {
+                    seeds.Add(ti);
+                }
+
+                base.MergeSourceInput (source, mergeType);
+
+                // Add the currently suggested songs to the played list, so
+                // they (and all added songs) do not get removed when the
+                // playlist is updated
+                played.AddRange(suggested);
+                played.AddRange(seeds);
+
+                List<DatabaseTrackInfo> exclude = new List<DatabaseTrackInfo>();
+                exclude.AddRange(played);
+                exclude.AddRange(skipped);
+
+                SimilarTracks (seeds, exclude, true);
             }
         }
         
@@ -246,8 +283,13 @@ namespace Banshee.Mirage
         private void Clear ()
         {
             current_track = 0;
+
             RemoveTrackRange ((DatabaseTrackListModel)TrackModel, new Hyena.Collections.RangeCollection.Range (0, Count));
-            tracksOverride.Clear ();
+
+            skipped.Clear ();
+            played.Clear ();
+            suggested.Clear ();
+
             Reload ();
         }
 
@@ -264,14 +306,68 @@ namespace Banshee.Mirage
             }
             
             ToggleAction action = (ToggleAction)uia_service.GlobalActions["ClearMiragePlaylistOnQuitAction"];
-            MiragePlugin.ClearOnQuitSchema.Set (action.Active);
+            MirageConfiguration.ClearOnQuitSchema.Set (action.Active);
         }
 
-        public void OnNextPrevAction(object o, EventArgs e)
-        {
-            skipped.Add(ServiceManager.PlayerEngine.CurrentTrack as DatabaseTrackInfo);
+        private class MiragePlaylistSource : PlaylistSource {
+
+            public MiragePlaylistSource() : 
+                base(AddinManager.CurrentLocalizer.GetString ("Mirage Playlist"), ServiceManager.SourceManager.MusicLibrary)
+            {
+            }
+
+            public void AddTracks (DatabaseTrackListModel from)
+            {
+                AddTrackRange(from, new Hyena.Collections.RangeCollection.Range(0, from.Count-1));
+                OnTracksAdded();
+            }
         }
-        
+
+        private void OnSavePlaylist (object o, EventArgs args)
+        {
+            MiragePlaylistSource playlist = new MiragePlaylistSource ();
+            playlist.Save ();
+            playlist.PrimarySource.AddChildSource (playlist);
+            playlist.AddTracks (DatabaseTrackModel);
+            playlist.NotifyUser ();
+
+            //SourceView.BeginRenameSource (playlist);
+        }
+
+        /*
+        private void OnBanFromPlaylist (object o, EventArgs args)
+        {
+        }
+        */
+
+        protected override void OnTracksRemoved ()
+        {
+            if (ServiceManager.SourceManager.ActiveSource != this) {
+                return;
+            }
+
+            // add removed tracks to exclude list
+            foreach (int track_id in TrackModel.Selection) {
+                DatabaseTrackInfo ti = (DatabaseTrackInfo)DatabaseTrackModel[track_id];
+
+                if (!played.Exists( delegate (DatabaseTrackInfo t) { return t.TrackId == ti.TrackId; })) {
+                    Log.DebugFormat ("Mirage - Adding {0}-{1} to exclude list",
+                        ti.TrackId, ti.TrackTitle);
+                    skipped.Add (ti);
+                }
+            }
+            Reload ();
+        }
+
+        private bool ContainsTrack(List<DatabaseTrackInfo> tlist, DatabaseTrackInfo ti)
+        {
+            return tlist.Exists(
+                delegate (DatabaseTrackInfo t)
+                {
+                    return t.TrackId == ti.TrackId;
+                });
+        }
+       
         private void Refresh ()
         {
             processed = ServiceManager.PlayerEngine.CurrentTrack as DatabaseTrackInfo;
@@ -287,20 +383,21 @@ namespace Banshee.Mirage
                     // We're playing another source
                     return;
                 }
+                played.Add(processed);
+
+                seeds.Clear();
                 seeds.Add(processed);
                 
-                List<DatabaseTrackInfo> skip = new List<DatabaseTrackInfo>();
-                
-                skip.AddRange(seeds);
-                if (skipped.Count > 0)
-                    skip.AddRange(skipped);
+                List<DatabaseTrackInfo> exclude = new List<DatabaseTrackInfo>();
+                exclude.AddRange(played);
+                exclude.AddRange(skipped);
 
-                SimilarTracks(seeds, skip);
+                SimilarTracks(seeds, exclude, false);
             }
         }
 
         private void SimilarTracks(List<DatabaseTrackInfo> tracks,
-                List<DatabaseTrackInfo> exclude)
+                List<DatabaseTrackInfo> exclude, bool append)
         {
             int[] trackIds = new int[tracks.Count];
             for (int i = 0; i < tracks.Count; i++) {
@@ -314,11 +411,8 @@ namespace Banshee.Mirage
                 excludeTrackIds[i] = exclude[i].TrackId;
             }
 
-            SimilarityCalculator sc =
-                    new SimilarityCalculator(trackIds, excludeTrackIds,
-                            db, UpdatePlaylist, 5);
-            Thread similarityCalculatorThread =
-                    new Thread(new ThreadStart(sc.Compute));
+            SimilarityCalculator sc = new SimilarityCalculator(trackIds, excludeTrackIds, db, UpdatePlaylist, append);
+            Thread similarityCalculatorThread = new Thread(new ThreadStart(sc.Compute));
             similarityCalculatorThread.Start();
         }
 
@@ -341,10 +435,18 @@ namespace Banshee.Mirage
                 case PlayerEvent.Iterate:
                     // if more than 60% of a track is played, use this track as
                     // a seed song for the next tracks.
+                   
                     if ((processed != ServiceManager.PlayerEngine.CurrentTrack) &&
                             (ServiceManager.PlayerEngine.Position > 
                              ServiceManager.PlayerEngine.Length * 0.6)) {
-                        Refresh ();
+
+                        // only refresh if the track was not played yet
+                        if (!ContainsTrack(played, ServiceManager.PlayerEngine.CurrentTrack as DatabaseTrackInfo)) {
+                            Refresh ();
+                        } else {
+                            processed = ServiceManager.PlayerEngine.CurrentTrack as DatabaseTrackInfo;
+                        }
+
                     }
                     break;
             }
@@ -352,16 +454,24 @@ namespace Banshee.Mirage
 
         bool IBasicPlaybackController.First ()
         {
-            return ((IBasicPlaybackController)this).Next (false);
+            ServiceManager.PlayerEngine.OpenPlay(GetTrack(current_track));
+            return true;
         }
 
         bool IBasicPlaybackController.Next (bool restart)
         {
+            if (ServiceManager.PlayerEngine.CurrentTrack == null) {
+                ServiceManager.PlayerEngine.OpenPlay(GetTrack(current_track));
+                return true;
+            }
+
             // We get the next track before removing the current one
             DatabaseTrackInfo next_track = NextTrack;
-            if (processed != ServiceManager.PlayerEngine.CurrentTrack) {
+            if ((processed != ServiceManager.PlayerEngine.CurrentTrack) &&
+                    (!ContainsTrack(played, ServiceManager.PlayerEngine.CurrentTrack as DatabaseTrackInfo))) {
                 RemovePlayingTrack ();
             }
+
             current_track++;
             ServiceManager.PlayerEngine.OpenPlay (next_track);
             return true;
@@ -380,6 +490,9 @@ namespace Banshee.Mirage
         {
             DatabaseTrackInfo playing_track = GetTrack (current_track);
             if (playing_track != null) {
+                skipped.Add(playing_track);
+                Log.DebugFormat ("Mirage - Adding {0}-{1} to exclude list (skipped)",
+                    playing_track.TrackId, playing_track.TrackTitle);
                 RemoveTrack (playing_track);
             }
         }
